@@ -74,6 +74,43 @@ impl Connection {
         self.read_and_deserialize(len as usize).await
     }
 
+    fn try_read_and_deserialize<'a, T>(&'a mut self, size: usize) -> Result<Option<T>>
+    where
+        T: Deserialize<'a>,
+    {
+        self.buffer.resize(size, 0);
+        match self.raw_conn.try_read(&mut self.buffer)? {
+            Some(_) => Ok(Some(from_bytes(&self.buffer)?)),
+            None => Ok(None),
+        }
+    }
+
+    fn try_read_header(&mut self) -> Result<Option<u32>> {
+        self.try_read_and_deserialize(4)
+    }
+
+    /// If it is readable, then the entire packet can be read in blocking manner.
+    ///
+    /// While this is indeed a blocking call, it is unlikely to block since ssh mux
+    /// master most likely would send it using one write/send.
+    ///
+    /// Even if it does employ multiple write/send, these functions would just return
+    /// immediately since the buffer for the unix socket is empty and should be big
+    /// enough for one message.
+    ///
+    /// If it is not readable, then it would return Ok(None).
+    fn try_read_response(&mut self) -> Result<Option<Response>> {
+        let len = match self.try_read_header()? {
+            Some(len) => len as usize,
+            None => return Ok(None),
+        };
+        loop {
+            if let Some(response) = self.try_read_and_deserialize(len)? {
+                break Ok(Some(response));
+            }
+        }
+    }
+
     fn get_request_id(&mut self) -> u32 {
         let request_id = self.request_id.0;
         self.request_id += Wrapping(1);
@@ -161,7 +198,7 @@ impl Connection {
         })
         .await?;
         for fd in fds {
-            self.raw_conn.send_with_fds(&[0], &[*fd])?;
+            self.raw_conn.send_with_fds(&[*fd]).await?;
         }
 
         let session_id = match self.read_response().await? {
@@ -414,7 +451,7 @@ mod tests {
         conn: Connection,
         cmd: &str,
     ) -> (EstablishedSession, (PipeWrite, PipeRead)) {
-        let session = Session::builder().cmd(cmd).build();
+        let session = Session::builder().cmd(cmd.into()).build();
 
         // pipe() returns (PipeRead, PipeWrite)
         let (stdin_read, stdin_write) = pipe().unwrap();
@@ -450,30 +487,30 @@ mod tests {
         assert_matches!(
             session_status,
             SessionStatus::Exited { exit_value, .. }
-                if exit_value == 0
+                if exit_value.unwrap() == 0
         );
     }
     run_test!(test_unordered_open_new_session, test_open_new_session_impl);
 
     async fn test_remote_socket_forward_impl(mut conn: Connection) {
-        let path = "/tmp/openssh-remote-forward.socket";
+        let path = Path::new("/tmp/openssh-remote-forward.socket");
 
         let output_listener = TcpListener::bind(("127.0.0.1", 1234)).await.unwrap();
 
         eprintln!("Requesting port forward");
         conn.request_port_forward(
             ForwardType::Remote,
-            &Socket::UnixSocket { path },
+            &Socket::UnixSocket { path: path.into() },
             &Socket::TcpSocket {
-                port: NonZeroU32::new(1234).unwrap(),
-                host: "127.0.0.1",
+                port: 1234,
+                host: "127.0.0.1".into(),
             },
         )
         .await
         .unwrap();
 
         eprintln!("Creating remote process");
-        let cmd = format!("/usr/bin/socat OPEN:/data,rdonly UNIX-CONNECT:{}", path);
+        let cmd = format!("/usr/bin/socat OPEN:/data,rdonly UNIX-CONNECT:{:#?}", path);
         let (established_session, stdios) = create_remote_process(conn, &cmd).await;
 
         eprintln!("Waiting for connection");
@@ -497,7 +534,7 @@ mod tests {
         assert_matches!(
             session_status,
             SessionStatus::Exited { exit_value, .. }
-                if exit_value == 0
+                if exit_value.unwrap() == 0
         );
     }
     run_test!(
@@ -506,10 +543,10 @@ mod tests {
     );
 
     async fn test_local_socket_forward_impl(conn0: Connection, mut conn1: Connection) {
-        let path = "/tmp/openssh-local-forward.socket";
+        let path = Path::new("/tmp/openssh-local-forward.socket").into();
 
         eprintln!("Creating remote process");
-        let cmd = format!("socat -u OPEN:/data UNIX-LISTEN:{} >/dev/stderr", path);
+        let cmd = format!("socat -u OPEN:/data UNIX-LISTEN:{:#?} >/dev/stderr", path);
         let (established_session, stdios) = create_remote_process(conn0, &cmd).await;
 
         sleep(Duration::from_secs(1)).await;
@@ -519,8 +556,8 @@ mod tests {
             .request_port_forward(
                 ForwardType::Local,
                 &Socket::TcpSocket {
-                    port: NonZeroU32::new(1235).unwrap(),
-                    host: "127.0.0.1",
+                    port: 1235,
+                    host: "127.0.0.1".into(),
                 },
                 &Socket::UnixSocket { path },
             )
@@ -546,7 +583,7 @@ mod tests {
         assert_matches!(
             session_status,
             SessionStatus::Exited { exit_value, .. }
-                if exit_value == 0
+                if exit_value.unwrap() == 0
         );
     }
     run_test2!(
@@ -554,6 +591,41 @@ mod tests {
         test_local_socket_forward_impl
     );
 
+    async fn test_unordered_try_wait_impl(conn: Connection) {
+        use TryWaitSessionStatus::{Exited, InProgress};
+
+        let (mut established_session, stdios) = create_remote_process(conn, "cat").await;
+
+        // cat would not exit until stdin is closed
+        for _i in 0..100 {
+            match established_session.try_wait().unwrap() {
+                InProgress(session) => established_session = session,
+                _ => panic!("Unexpected return value from try_wait"),
+            }
+        }
+
+        drop(stdios);
+
+        // Wait for the remote process to exit
+        sleep(Duration::from_millis(100)).await;
+
+        // try_wait for three times
+        for _i in 0..3 {
+            match established_session.try_wait().unwrap() {
+                InProgress(session) => established_session = session,
+                Exited {
+                    exit_value: _exit_value,
+                } => {
+                    return;
+                }
+                _ => panic!("Unexpected return value from try_wait"),
+            }
+        }
+
+        panic!("try_wait timed out");
+    }
+    run_test!(test_unordered_try_wait, test_unordered_try_wait_impl);
+
     async fn test_request_stop_listening_impl(mut conn: Connection) {
         conn.request_stop_listening().await.unwrap();
 
